// +=============================================+
// |  ____  _            _     ____ _ _          |
// | | __ )| | ___   ___| | __/ ___(_) |_ _   _  |
// | |  _ \| |/ _ \ / __| |/ / |   | | __| | | | |
// | | |_) | | (_) | (__|   <| |___| | |_| |_| | |
// | |____/|_|\___/ \___|_|\_\\____|_|\__|\__, | |
// |                                      |___/  |
// +=============================================+
// | This is the Packaged Functions of the script|
// | Do not modify this file.			 |
// +---------------------------------------------+

package blockCity
{
	// +---------------------------------------------------+
	// | This places the joining person into the minigame. |
	// +---------------------------------------------------+
	function GameConnection::onClientEnterGame(%client) 
	{
		Parent::onClientEnterGame(%client);

		if(!BCDB.ExistUser(%client.BL_ID))
			BCDB.AddUser(%client.BL_ID);

		if(isVIP(%client.BL_ID) && !(%client.isAdmin || %client.isSuperAdmin))
		{
			messageClient(%client, "", "\c3" @ "You are a VIP on this server!");
			messageClient(%client, "", "\c3" @ "You are able to use the Wrench!");
			%client.clanSuffix="<color:FCD116> [VIP]";
		}
		if(%client.isAdmin || %client.isSuperAdmin)
			%client.clanSuffix="<color:EE0000> [Server Mod]";
		if(!isObject($DediMini))
			CheckDediMini();
		$DediMini.addMember(%client);
		
		if($Pref::blockCity::enableDM)
			messageClient(%client, "","\c4" @ "Type" @ "\c6" SPC "/DM" SPC "\c4" @ "to start Deathmatching!");

		centerPrint(%client,"<bitmap:Add-Ons/Script_BlockCity/ssblocklandcity>", 5);

		if($Pref::BlockCity::AutoDisplayHelp)
			serverCmdHelp(%client);
	}

	// +--------------------------------------------------+
	// | This shuts down a non-dedicated server properly. |
	// +--------------------------------------------------+
	function disconnectedCleanup()
	{
		for(%i=0;%i<ClientGroup.getCount();%i++) 
		{
			%cl=ClientGroup.getObject(%i);
			if(%cl.miniGame.getID() == $DediMini.getID())
			{
				$DediMini.removeMember(%cl);
			}
			else if(%cl.miniGame.getID() == $DMMini.getID())
			{
				$DMMini.removeMember(%cl);
			}

		}
		$DediMini.schedule(0,"delete");
		$DMMini.schedule(0,"delete");
		$Temp::ServerAutoMinigameRanOnce = false;

		//SimpleDB
		BCDB.saveData();
		BCDB.delete();		
		//End SimpleDB


		Parent::disconnectedCleanup();
	}

	// +--------------------------------------------------------------+
	// | This restricts players from being able to create a minigame. |
	// +--------------------------------------------------------------+
	function serverCmdCreateMinigame(%client,%a,%b,%c,%d,%e,%f,%g) 
	{
		if(isObject($DediMini)) 
			messageClient(%client,'',"\c0Sorry, you cannot leave the default minigame.");
	}

	// +---------------------------------------------------+
	// | This restricts players from leaving the minigame. |
	// +---------------------------------------------------+
	function serverCmdLeaveMinigame(%client) 
	{
		if(isObject($DediMini)) 
			messageClient(%client,'',"\c0Sorry, you cannot leave the default minigame.");
	}
	
	// +--------------------------------------------------------+
	// | This is used if spawn points are used in the minigame. |
	// +--------------------------------------------------------+
	function getBrickSpawnPoint(%client) 
	{
		if(%client !$="")
			Parent::getBrickSpawnPoint(%client);
	}
	
	// +-------------------------------------------------+
	// | These Restrictions prevent resource exploiting. |
	// +-------------------------------------------------+
	// Now with Selective Disabling
	function FxDTSBrick::SpawnProjectile(%brick,%b,%projectile,%d,%e)
	{
		if($Pref::NoProjectile)
			return;
		if($Pref::BlockCity::SelectiveProjectiles && %projectile.disallowSpawn)
			return;
		parent::SpawnProjectile(%brick,%b,%projectile,%d,%e);
	}

	// +-------------------------------------------------+
	// | These Restrictions prevent resource exploiting. |
	// +-------------------------------------------------+
	// Now With Selective Projectile Disabling
	function player::SpawnProjectile(%a,%b,%projectile,%d,%e)
	{
		if($Pref::NoPlayerProjectile)
			return;
		if($Pref::BlockCity::SelectiveProjectiles && %projectile.disallowSpawn)
			return;
		parent::SpawnProjectile(%a,%b,%projectile,%d,%e);
	}
	
	// +----------------------------------------------------------------------+
	// | This modifies the wrench so that it can only be used by VIP members. |
	// +----------------------------------------------------------------------+
	function wrenchImage::onHitObject(%this, %player, %a, %brick, %particle, %f)
	{
		if(%brick.getClassName() !$= "fxDtsBrick") return;
		if(%brick.getDatablock().isMDRMResource)
		{
			%player.client.centerprint("<just:center><color:FFBBBB>You cannot modify a resource brick!",2);
			return;
		}
		if((!$Pref::blockCity::enableVIP || isVIP(%player.client.BL_ID)) || %player.client.isAdmin)
		{
			parent::onHitObject(%this, %player, %a, %brick, %particle, %f);
			return;
		}
		%player.client.centerprint("<just:center><color:FFBBBB>Only VIP members can use Wrench features!",2);
	}

	// +------------------------------------------------------------+
	// | This stops users from being able to paint resource bricks! |
	// +------------------------------------------------------------+
	function paintProjectile::OnCollision(%a, %player, %brick, %b, %c, %d)
	{
		// If %brick is not an object, end.
		if(!isObject(%brick)) return;
		if(%brick.getClassName() $= "fxDtsBrick")
		{
			if(%brick.getDatablock().isMDRMResource && !%player.client.isAdmin)
			{
				%player.client.centerprint("<just:center><color:FFBBBB>You cannot paint a resource brick!",2);
				return;
			}
		}
		parent::OnCollision(%a, %player, %brick, %b, %c, %d);
	}
	
	// +--------------------------------------------------------+
	// | This packaged function checks what bricks are planted. |
	// +--------------------------------------------------------+
	function fxDtsBrick::onPlant(%brick)
	{
		parent::onPlant(%brick);

		// +-------------+
		// | Trust Check |
		// +-------------+
		if(isObject(%brick.getDownBrick(0)))
			%trust1 = getTrustLevel(%brick,%brick.getDownBrick(0));
		else
			%trust1 = 4;
		if(isObject(%brick.getUpBrick(0)))
			%trust2 = getTrustLevel(%brick,%brick.getUpBrick(0));
		else
			%trust2 = 4;
		if(%trust1 < 1 || %trust2 < 1)
		{
			if($BlockCity::Debug) warn("[BlockCity]::OnPlant() - TRUST CHECK LEVELS: BrickBelow: "@%trust1@" BrickAbove:"@%trust2@"     --- 4 = NoBrick");
			%brick.schedule(0, "delete");
			return false;
		}

		if(!getBrickGroupFromObject(%brick).client.isAdmin)
			%brickCheck = plantBrickCheck(%brick,getBrickGroupFromObject(%brick).client);
		else
			%brickCheck = 1;
		if(%brickCheck)
		{
			if($BlockCity::Debug) warn("[BlockCity]::OnPlant() -- OK");
			if(%brick.getDatablock().isMDRMResource)
			{
				%brick.setColor(getClosestPaintColor(%brick.getDatablock().MDRM_Color));
				regrow(%brick);
			}
		} else
		{
			if($BlockCity::Debug) warn("[BlockCity]::OnPlant() -- DENY");
			%brick.schedule(0, "delete");
			return;
		}
	}

	//Here we determine if they COULD plant the brick, but we do not charge.
	// If there is no way they can plant brick we stop them.

	function serverCmdPlantBrick(%client)
	{
		%brick = %client.Player.tempBrick;  // This is the ghost brick, but it has the same datablock so it works.
		%data = BrickTest(%brick);

		if($BlockCity::Debug) warn("ServerCmdPlantBrick -- Allow Brick Plant?:"@%data);

		if(!%data && !%client.isAdmin)
			return false;
		else
			parent::serverCmdPlantBrick(%client);
	}

	// +-------------------+
	// | REFUNDS SYSTEM v2 |
	// +-------------------+
	function fxDTSBrick::onDeath(%brick)
	{
		%client = getbrickgroupfromobject(%brick).client;
		if((%brick.isBasePlate || %brick.getDatablock().isMDRMResource) && !%client.isAdmin)
		{
			//if(%brick.isBaseplate)
				//givePlayer(%client,100,"Refund");
			if(%brick.getDatablock().isMDRMResource)
				givePlayer(%client,mFloor(%brick.getDatablock().MDRM_BPH*%brick.getDatablock().MDRM_TotalHits*75),"Refund");
		}

		Parent::onDeath(%brick);
	}

	function GameConnection::onDeath(%Client, %b, %Killer, %d, %e)
	{
		parent::onDeath(%Client, %b, %Killer, %d, %e);
		
		if(!(%Killer == %Client) && !(%Killer == 0))
		{
			givePlayer(%Killer,$Pref::BlockCity::DMWinReward,"DM Kill");
			chargePlayer(%Client,$Pref::BlockCity::DMLossReward,"DM Loss",1);
		}
	}

	function serverCmdDuplicator(%client)
	{
		if(!isfile("Add-Ons/Tool_Duplicator/server.cs"))
			return;
		
		if($Pref::Duplicator::AdminOnly && (!%client.isAdmin && !%client.isSuperAdmin))
			return;

		if($Pref::blockCity::enableVIP &&(!isVIP(%client.bl_ID) && !%client.isadmin))
		{
			messageclient(%client, "","You must be a VIP member to use the Duplicator Wand!");
			return;
		}
		parent::serverCmdDuplicator(%client);
	}

};
activatePackage(blockCity);

